from collections import deque
from typing import Set

from .job_graph import JobGraph


class Scheduler:
    job: JobGraph
    finished: Set[str] = set()
    candidates: deque[str] = deque()

    is_finished: bool = False

    def __init__(self, job: JobGraph):
        self.job = job

        self.commit(self.job.begin_task)

    def commit(self, step: str):
        """declares a step has finished.
        Args:
            step (str): the id of the step.
        """

        if step in self.finished:
            return

        self.finished.add(step)

        if step in self.candidates:
            self.candidates.remove(step)

        for next_step in self.job.next[step]:
            if self.job.prev[next_step] - self.finished:
                continue  # The step is not ready to run.
            if next_step in self.candidates:
                continue  # Already added to candidates.
            if next_step in self.finished:
                continue
            self.candidates.append(next_step)

    def schedule(self) -> str | None:
        """schedule a step that is able to be executed.

        Returns:
            str | None: schedule a task or if no task is available, return None.
        """
        if len(self.candidates) != 0:
            step = self.candidates.popleft()
            if step == self.job.end_task:
                self.is_finished = True
                return None
            return step
        else:
            return None


if __name__ == "__main__":
    taskgraph = JobGraph("greate")
    taskgraph.add_edge("1", "2")
    taskgraph.add_edge("1", "3")
    taskgraph.add_edge("2", "4")
    taskgraph.add_edge("2", "5")
    taskgraph.add_edge("3", "5")
    taskgraph.add_edge("3", "6")
    taskgraph.add_edge("4", "7")
    taskgraph.add_edge("5", "7")
    taskgraph.add_edge("6", "7")
    scheduler = Scheduler(taskgraph)

    while True:
        step = scheduler.schedule()
        if step == None:
            break
        scheduler.commit(step)
        print(step)



import asyncio
import logging
import logging.handlers
import multiprocessing as mp
import typing as t

from .job_graph import JobGraph

STOP = "__STOP__"

logger = logging.getLogger(__name__)


def worker(
    fn: t.Callable[[dict], dict],
    src_queues: dict[str, mp.Queue],
    dest_queues: dict[str, mp.Queue],
    log_queue: mp.Queue,
):
    qh = logging.handlers.QueueHandler(log_queue)

    logger.addHandler(qh)
    logger.setLevel(logging.DEBUG)

    while True:
        src_datas = {task: queue.get() for task, queue in src_queues.items()}
        if STOP in src_datas.values():
            for dest, queue in dest_queues.items():
                queue.put(STOP)
            return

        dest_datas = fn(src_datas)

        assert dest_datas.keys() == dest_queues.keys()

        logger.debug(
            f"prev: {list(src_queues.keys())}\tnext: {list(dest_queues.keys())}\tsrc_data: {src_datas}\tdest_data: {dest_datas}"
        )
        for dest, data in dest_datas.items():
            dest_queues[dest].put(data)


class MPExecutor:
    """Multi-Progessing Executor
    """

    _job: JobGraph
    _task_fn: dict[str, t.Callable[[dict], dict]]

    _src_queues: dict[str, dict[str, mp.Queue]]
    _dest_queues: dict[str, dict[str, mp.Queue]]
    _processes: list[mp.Process]

    def __init__(self, job: JobGraph, task_fn: dict[str, t.Callable[[dict], dict]]):
        """_summary_

        Args:
            job (JobGraph): job is described as a digraph, the digraph is then traversed and queues between each task is created.
            task_fn (dict[str, t.Callable[[dict], dict]]): task function sould accept a dict mapping input taskname to input data, and return a dict mapping output taskname to output data.
        """
        self._job = job
        self._task_fn = task_fn

        # initialize (src/dest)-queue mapping for each task.
        self.__init_queues()

        # initialize processes
        self.__init_processes()

    def __init_queues(self):
        tasks = [self._job.begin_task] + self._job.serialize()

        self._src_queues: dict[str, dict[str, mp.Queue]] = {}
        self._dest_queues: dict[str, dict[str, mp.Queue]] = {}

        for task in tasks + [self._job.end_task]:
            self._src_queues[task] = {}
            self._dest_queues[task] = {}

        for task in tasks:
            dest_set = self._job.next[task]
            for dest in dest_set:
                q = mp.Queue()
                self._dest_queues[task][dest] = q
                self._src_queues[dest][task] = q

    def __init_processes(self):
        logging_queue = mp.Queue()

        listener = logging.handlers.QueueListener(logging_queue, *logger.handlers)
        listener.start()

        self._processes = [
            mp.Process(
                target=worker,
                args=(
                    self._task_fn[task],
                    self._src_queues[task],
                    self._dest_queues[task],
                    logging_queue,
                ),
            )
            for task in self._job.node - set([self._job.begin_task, self._job.end_task])
        ]

        for p in self._processes:
            p.start()

    def run(self, input: dict[str, t.Any]) -> dict[str, t.Any]:
        """run on the input and block until all the output task generate results.

        Args:
            input (dict[str, Any]): a dictionary that map input task name to the input data

        Returns:
            dict[str, Any]: a dictionary that map output task name to the output data.
        """
        for task, value in input.items():  # a graph might have multiple inputs.
            self._dest_queues[self._job.begin_task][task].put(value)

        if STOP in input.values():
            return {}

        def __print_get(task: str, queue: mp.Queue) -> t.Any:
            ret = queue.get()
            logger.debug(f"GETTING {task}: {ret}")
            return ret

        result = {
            task: __print_get(task, queue)
            for task, queue in self._src_queues[self._job.end_task].items()
        }

        return result

    async def arun(self, input: dict[str, t.Any]) -> dict[str, t.Any]:
        """async version of `run`

        Args:
            input (dict[str, t.Any]): a dictionary that map input task name to the input data

        Returns:
            Coroutine: a coroutine that wrap output of `run`.
        """
        return await asyncio.to_thread(self.run, input)

    def shutdown(self):
        """join all the subprocesses."""
        input_tasks = self._job.next[self._job.begin_task]
        inputs = {task: STOP for task in input_tasks}
        self.run(inputs)
        for p in self._processes:
            p.join()



from collections import deque
from typing import Dict, List, Set

import graphviz

from .task import Task


class JobGraph:
    """
    A Special Digraph that is always connected.
    """

    node: Set[str]
    next: Dict[str, set[str]]
    prev: Dict[str, set[str]]

    def __init__(self, name: str):
        self.name = name

        self.node = set()
        self.next = dict()
        self.prev = dict()

        self.begin_task = "__begin__"
        self.end_task = "__end__"

        self.node.add(self.begin_task)
        self.node.add(self.end_task)

        self.prev[self.begin_task] = set()
        self.next[self.begin_task] = set()
        self.prev[self.end_task] = set()
        self.next[self.end_task] = set()

    def __link_nodes(self, task1: str, task2: str):
        self.next[task1].add(task2)
        self.prev[task2].add(task1)

    def __unlink_nodes(self, task1: str, task2: str):
        self.next[task1].remove(task2)
        self.prev[task2].remove(task1)

    def add_node(self, task: str):
        if task in self.node:
            return

        self.node.add(task)

        self.next[task] = set()
        self.prev[task] = set()

        # create default links
        self.__link_nodes(self.begin_task, task)
        self.__link_nodes(task, self.end_task)

    def add_edge(self, task1: str, task2: str):

        if task1 not in self.node:
            self.add_node(task1)
        if task2 not in self.node:
            self.add_node(task2)

        # Remove default edges
        if self.end_task in self.next[task1]:
            self.__unlink_nodes(task1, self.end_task)

        if self.begin_task in self.prev[task2]:
            self.__unlink_nodes(self.begin_task, task2)

        self.__link_nodes(task1, task2)

        return self

    def serialize(self) -> List[str]:
        serialization = []
        visited = set()
        dependencies = deque([self.end_task])
        while dependencies:
            head = dependencies.popleft()
            if head in visited:
                continue
            serialization.append(head)
            visited.add(head)

            for d in self.prev[head]:
                if d not in visited:
                    dependencies.append(d)

        serialization.remove(self.begin_task)
        serialization.remove(self.end_task)
        serialization.reverse()
        return serialization

    def draw_graph(self, render: bool = False, filename=None):
        """Draw graph using graphviz.

        Args:
            render (bool, optional): whether to render the graph. Defaults to False.
            filename (_type_, optional): The path to store the rendered graph. Defaults to None.

        Returns:
            graphviz.Digraph: _description_
        """
        dot = graphviz.Digraph(comment=self.name)

        for node in self.node:
            dot.node(node)

        for node, edges in self.next.items():
            for edge in edges:
                dot.edge(node, edge)

        if render:
            if filename is None:
                filename = f"JobGraph_{self.name}"
            dot.render(filename, view=False)
        return dot


# Example usage
if __name__ == "__main__":

    graph_task = JobGraph("anonymous")

    task1 = Task(lambda: print("task 1"), name="task1")
    task2 = Task(lambda: print("task 2"), name="task2")
    task3 = Task(lambda: print("task 3"), name="task3")

    graph_task.add_edge(task1.name, task2.name)  # task1 -> task2
    graph_task.add_edge(task2.name, task3.name)  # task2 -> task3
    graph_task.add_edge(task1.name, task3.name)

    dot = graph_task.draw_graph(render=True)

    print(graph_task.serialize())
    print(repr(task1))
    print(task1)



class Task:
    """
    Encapsulates a function call with its arguments and keyword arguments,
    allowing it to be executed as a single callable entity.

    Attributes:
        name (str): The name of the step.

    """

    def __init__(self, func, args: tuple = (), kwargs: dict = {}, name=None):
        """
        Args:
            func (callable): The function to be called.
            args (tuple, optional): The positional arguments to be passed to the function. Defaults to ().
            kwargs (dict, optional): The keyword arguments to be passed to the function. Defaults to {}.
            name (str, optional): The name of the step. If not provided, a default name is generated using the function's name and the hash of the instance.
        """
        self._func = func
        self._args = args
        self._kwargs = kwargs

        self.name = name if name is not None else f"{func.__name__}_{str(hash(self))}"

    def __call__(self):
        return self._func(*self._args, **self._kwargs)

    def __str__(self) -> str:
        return 'Step "self.name"'

    def __repr__(self) -> str:
        return f"_Step(func={self._func!r}, args={self._args!r}, kwargs={self._kwargs!r}, name={self.name!r})"



import datetime
import logging
import os
import typing as t
from random import random

import graphviz
import jsonlines
from pydantic import BaseModel

from .job_graph import JobGraph
from .scheduler import Scheduler
from .task import Task

logger = logging.getLogger(__name__)


class _Record(BaseModel):
    task_name: str
    timestamp: str = datetime.datetime.now().strftime(r"%d/%m/%Y,%H:%M:%S")


class Executor:
    scheduler: Scheduler
    tasks: t.Dict[str, Task]
    finished_tasks: t.List[_Record] = []

    def __init__(self, scheduler, tasks, cache_fname=None) -> None:
        self.scheduler = scheduler
        self.tasks = tasks

        self.__dot = self.scheduler.job.draw_graph(render=False)

        self.cache_fname = (
            cache_fname if cache_fname is not None else self.__get_cache_fname()
        )
        os.makedirs(os.path.dirname(self.cache_fname), exist_ok=True)

        if os.path.exists(self.cache_fname):
            self.__load_cache()

    def __load_cache(self):
        with jsonlines.open(self.cache_fname, "r") as reader:
            for line in reader:
                record = _Record(**line)
                self.finished_tasks.append(record)
                scheduler.commit(record.task_name)
                self.__dot.node(record.task_name, color="green")

    def __add_cache(self, task_name):
        record = _Record(task_name=task_name)
        self.finished_tasks.append(record)
        with jsonlines.open(self.cache_fname, "a") as writer:
            writer.write(record.model_dump())

    def __get_cache_fname(self):
        return os.path.join("_cache", "resume", self.scheduler.job.name + ".jsonl")

    def run(self):
        while True:
            task_name = self.scheduler.schedule()
            if self.scheduler.is_finished:
                break
            assert task_name is not None
            self.__dot.node(task_name, color="red")

            self.tasks[task_name]()
            self.scheduler.commit(task_name)

            self.__add_cache(task_name)
            self.__dot.node(task_name, color="green")

    def draw_graph(self) -> graphviz.Digraph:
        """Draw the graph in graphviz.Digraph.

        The graph is updated as execution goes. Green indicates success,
        Red means failure.

        Returns:
            graphviz.Digraph: the graph.
        """
        return self.__dot


if __name__ == "__main__":
    job = JobGraph("greate")
    job.add_edge("1", "2")
    job.add_edge("1", "3")
    job.add_edge("2", "4")
    job.add_edge("2", "5")
    job.add_edge("3", "5")
    job.add_edge("3", "6")
    job.add_edge("4", "7")
    job.add_edge("5", "7")
    job.add_edge("6", "7")
    job.draw_graph(render=True)
    scheduler = Scheduler(job)

    def madman(i):
        if random() < 0.2:
            raise RuntimeError
        print(i)

    tasks = {f"{i}": Task(lambda i=i: madman(i)) for i in range(1, 8)}

    executor = Executor(scheduler, tasks)
    try:
        executor.run()
    except:
        pass
    finally:
        graph = executor.draw_graph()
        graph.render(f"JobGraph-{executor.scheduler.job.name}")
